#include "BehaviorEditor.h"

NodePresent::NodePresent(QWidget* pwidget, double& width_bind, std::shared_ptr<LogicalNode> bind)
	: _widget_p(pwidget), _node_bind(bind), _column_width(width_bind) {
}

QRectF NodePresent::contentMeasure() const
{
	auto metrics = this->_widget_p->fontMetrics();
	return metrics.boundingRect(this->_node_bind->rtName());
}

QRectF NodePresent::boundingRect() const
{
	auto rect = contentMeasure();
	rect.setWidth(std::max(_column_width, rect.width()));
	_column_width = rect.width();

	return rect + QMargins(0, 0, 30, 30);
}

#include <QStyleOptionGraphicsItem>
void NodePresent::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	painter->drawRect(boundingRect());
	painter->fillRect(option->rect - QMargins(8, 8, 8, 8), Qt::gray);

	painter->translate(QPointF(15, 15));
	auto rect = contentMeasure();
	painter->drawText(rect, 0, this->_node_bind->rtName());
}

BehaviorsPresent::BehaviorsPresent(QWidget* parent /*= nullptr*/)
	: QGraphicsView(parent) {
	this->setScene(&_bind_scene);

	auto font = this->font();
	font.setPixelSize(20);
	this->setFont(font);
}

void BehaviorsPresent::setRoot(std::shared_ptr<BehaviorMapNode> root)
{
	qDeleteAll(_present_peers);
	_present_peers.clear();

	this->_bind_maproot = root;
	relayout();
}

void BehaviorsPresent::presentAllocate(std::shared_ptr<LogicalNode> ins)
{
	switch (ins->nodeKind())
	{
	case NodeKind::MAPNODE:
	case NodeKind::PARALLELNODE:
	case NodeKind::SEQUENCENODE:
	case NodeKind::SELECTORNODE:
		if (ins->nodeKind() != NodeKind::MAPNODE || !ins->bindMap()) {
			auto type = ins->nodeKind();
			for (auto it : ins->children())
				presentAllocate(it);
		}

	default:
		if (_present_peers.contains(ins))
			return;

		for (auto idx = _column_aligns.size(); idx < ins->depth() + 1; ++idx)
			_column_aligns.append(0);

		auto& current_width = _column_aligns[ins->depth()];
		_present_peers[ins] = new NodePresent(this, current_width, ins);
		this->_bind_scene.addItem(_present_peers[ins]);
		break;
	}
}

void BehaviorsPresent::presentRelease(std::shared_ptr<LogicalNode> ins)
{
	switch (ins->nodeKind())
	{
	case NodeKind::MAPNODE:
	case NodeKind::PARALLELNODE:
	case NodeKind::SELECTORNODE:
	case NodeKind::SEQUENCENODE: {
		for (auto item : ins->children()) {
			presentRelease(item);
		}
	}break;
	default:
		if (this->_present_peers.contains(ins))
			delete this->_present_peers[ins];
		this->_present_peers.remove(ins);
		break;
	}
}

void BehaviorsPresent::relayout()
{
	if (!_bind_maproot)
		return;

	// ʼʾڵ
	presentAllocate(_bind_maproot);

	QHash<std::shared_ptr<LogicalNode>, std::pair<QSizeF, QSizeF>> results;
	// ߴ
	outlineMeasure(_bind_maproot, results);
	// ԪŲ
	nodeRelayout(results, _bind_maproot, QPointF());
}

QSizeF BehaviorsPresent::outlineMeasure(std::shared_ptr<LogicalNode> ins, QHash<std::shared_ptr<LogicalNode>, std::pair<QSizeF, QSizeF>>& _outline_occupy)
{
	QSizeF outline_box(0, 0);

	switch (ins->nodeKind()) {
	case NodeKind::MAPNODE:
	case NodeKind::PARALLELNODE:
	case NodeKind::SELECTORNODE:
	case NodeKind::SEQUENCENODE:
		// ӽڵ㼯ϳߴ磬ڵΪڵ
		if (ins->nodeKind() != NodeKind::MAPNODE || !ins->bindMap()) {
			for (auto nit : ins->children()) {
				auto nit_size = outlineMeasure(nit, _outline_occupy);
				outline_box.setWidth(std::max(outline_box.width(), nit_size.width() + _space_h));
				outline_box += QSizeF(0, nit_size.height());
			}
		}
	default: {
		auto root_peer = _present_peers[ins];
		auto root_box = root_peer->boundingRect();
		outline_box += QSizeF(root_box.width(), 0);
		outline_box.setHeight(std::max(outline_box.height(), root_box.height()));

		_outline_occupy[ins] = std::make_pair(outline_box, root_box.size());
		return outline_box;
	}break;
	}
}

QPointF BehaviorsPresent::nodeRelayout(QHash<std::shared_ptr<LogicalNode>, std::pair<QSizeF, QSizeF>>& _outline_occupy,
	std::shared_ptr<LogicalNode> ins, const QPointF& origin_offset) {
	// ȡߴ¼
	auto outline_occupy = _outline_occupy[ins];
	auto node_outline = outline_occupy.first;
	auto node_occupy = outline_occupy.second;

	// ȡʾڵ
	auto present_node = _present_peers[ins];

	// ýڵλ
	QPointF local_origin_offset(0, (node_outline.height() - node_occupy.height()) / 2);
	present_node->setPos(origin_offset + local_origin_offset);

	// ӽڵλ
	QPointF child_origin = origin_offset + QPointF(node_occupy.width() + _space_h, 0);
	switch (ins->nodeKind())
	{
	case NodeKind::ACTIONNODE:
	case NodeKind::COMPARENODE:
		break;
	default:
		// ӽڵ㼯λãڵǶ˸ڵ
		if (ins->nodeKind() != NodeKind::MAPNODE || !ins->bindMap()) {
			for (auto item : ins->children()) {
				child_origin = nodeRelayout(_outline_occupy, item, child_origin);
			}
		}break;
	}

	// ½λ
	return origin_offset + QPointF(0, node_outline.height());
}

uint qHash(const std::shared_ptr<LogicalNode> data, uint seed) noexcept
{
	return qHash((void*)data.get(), seed);
}

#include <QSplitter>
BehaviorEditor::BehaviorEditor(QWidget* parent /*= nullptr*/)
	: QMainWindow(parent),
	_type_view(new QListView(this)),
	_type_model(new QStandardItemModel(this)),
	_logical_present(new BehaviorsPresent(this)),
	_logs_present(new QTextBrowser(this))
{
	_glbal_loader = std::make_shared<MessageLoader>();
	_global_kernal = std::make_shared<MapKernal>(_glbal_loader);
	_global_kernal->initial();

	auto font = this->font();
	font.setPixelSize(20);
	_type_view->setFont(font);

	auto split_v = new QSplitter(Qt::Vertical, this);
	this->setCentralWidget(split_v);

	auto split_h = new QSplitter(Qt::Horizontal, this);
	split_v->addWidget(split_h);
	split_v->addWidget(_logs_present);

	split_h->addWidget(_logical_present);
	split_h->addWidget(_type_view);
	_type_view->setModel(_type_model);

	nodeTypesViewInit(_type_model);
}

#include <QStandardItem>
void BehaviorEditor::nodeTypesViewInit(QStandardItemModel* m)
{
	auto types = this->_global_kernal->nodeTypes();
	std::sort(types.begin(), types.end());
	for (auto type : types) {
		auto row_item = new QStandardItem(type);
		row_item->setEditable(false);
		m->appendRow(row_item);
	}
}
